@@ -23,8 +23,20 @@ from typing import Callable, TypeAlias
 
 import math
 
-import slidingpuzzle
-from slidingpuzzle.board import Board, freeze_board
+import numpy as np
+
+from slidingpuzzle.board import (
+    BLANK_TILE,
+    Board,
+    find_blank,
+    get_goal_tile,
+    get_goal_y,
+    get_goal_x,
+    get_goal_yx,
+    freeze_board,
+    is_solved,
+    swap_tiles,
+)
 
 
 Heuristic: TypeAlias = Callable[[Board], int | float]
@@ -44,14 +56,14 @@ def euclidean_distance(board: Board) -> float:
     Returns:
         The sum of all tile distances from their goal positions
     """
-    w = len(board[0])
+    h, w = board.shape
     dist = 0.0
     for y, row in enumerate(board):
         for x, tile in enumerate(row):
-            if slidingpuzzle.EMPTY_TILE == tile:
+            if BLANK_TILE == tile:
                 continue
-            a = abs(y - (tile - 1) // w)
-            b = abs(x - (tile - 1) % w)
+            goal_y, goal_x = get_goal_yx(h, w, tile)
+            a, b = abs(y - goal_y), abs(x - goal_x)
             dist += math.sqrt(a**2 + b**2)
     return dist
 
@@ -69,71 +81,80 @@ def hamming_distance(board: Board) -> int:
     Returns:
         The number of tiles out of their goal positions
     """
-    w = len(board[0])
+    h, w = board.shape
     dist = 0
     for y, row in enumerate(board):
         for x, tile in enumerate(row):
-            if slidingpuzzle.EMPTY_TILE == tile:
+            if BLANK_TILE == tile:
                 continue
-            orig_y = (tile - 1) // w
-            orig_x = (tile - 1) % w
-            if y != orig_y or x != orig_x:
+            goal_y, goal_x = get_goal_yx(h, w, tile)
+            if y != goal_y or x != goal_x:
                 dist += 1
     return dist
 
 
 def linear_conflict_distance(board: Board) -> int:
     r"""
-    Starts with Manhattan distance and adds an additional 2 for every out-of-place
-    pair of tiles in the same row or column that are also both in their target row
-    or column. It will take at least 2 additional moves to reshuffle the tiles into
-    their correct positions.
+    This is a simplified variant of the original Linear Conflict Distance.
+    It starts with Manhattan distance. Then for each row and column, the maximum number
+    of conflicts are identified, and 2 * this number is added to the distance.
+    (It will take at least 2 additional moves to reshuffle the conflicting tiles into
+    their correct positions.) This is an admissible improvement over Manhattan distance.
 
     Args:
         board: The board
 
     Returns:
-        A slightly more accurate heuristic value than Manhattan distance.
+        Estimated distance to goal.
     """
-    h = len(board)
-    w = len(board[0])
+    h, w = board.shape
     dist = manhattan_distance(board)
-    # check out-of-place cols in each row
+
+    # check out-of-place tiles in each row
     for y in range(h):
+        max_conflicts = 0
         for x1 in range(w):
-            if slidingpuzzle.EMPTY_TILE == board[y][x1]:
+            if BLANK_TILE == board[y, x1]:
                 continue
-            tile1 = board[y][x1]
-            target_row1 = (tile1 - 1) // w
+            conflicts = 0
+            tile1 = board[y, x1]
+            target_row1 = get_goal_y(h, w, tile1)
             if y != target_row1:
                 continue
             for x2 in range(x1 + 1, w):
-                if slidingpuzzle.EMPTY_TILE == board[y][x2]:
+                if BLANK_TILE == board[y, x2]:
                     continue
-                tile2 = board[y][x2]
-                target_row2 = (tile2 - 1) // w
+                tile2 = board[y, x2]
+                target_row2 = get_goal_y(h, w, tile2)
                 if y != target_row2:
                     continue
                 if target_row1 == target_row2 and tile2 < tile1:
-                    dist += 2
-    # check out-of-place rows in each col
+                    conflicts += 1
+            max_conflicts = max(max_conflicts, conflicts)
+        dist += 2 * max_conflicts
+
+    # check out-of-place tiles in each col
     for x in range(w):
+        max_conflicts = 0
         for y1 in range(h):
-            if slidingpuzzle.EMPTY_TILE == board[y1][x]:
+            if BLANK_TILE == board[y1, x]:
                 continue
-            tile1 = board[y1][x]
-            target_col1 = (tile1 - 1) % w
+            conflicts = 0
+            tile1 = board[y1, x]
+            target_col1 = get_goal_x(h, w, tile1)
             if x != target_col1:
                 continue
             for y2 in range(y1 + 1, h):
-                if slidingpuzzle.EMPTY_TILE == board[y2][x]:
+                if BLANK_TILE == board[y2, x]:
                     continue
-                tile2 = board[y2][x]
-                target_col2 = (tile2 - 1) % w
+                tile2 = board[y2, x]
+                target_col2 = get_goal_x(h, w, tile2)
                 if x != target_col2:
                     continue
                 if target_col1 == target_col2 and tile2 < tile1:
-                    dist += 2
+                    conflicts += 1
+            max_conflicts = max(max_conflicts, conflicts)
+        dist += 2 * max_conflicts
     return dist
 
 
@@ -152,15 +173,14 @@ def manhattan_distance(board: Board) -> int:
     Returns:
         The sum of all tile distances from their goal positions
     """
-    w = len(board[0])
+    h, w = board.shape
     dist = 0
     for y, row in enumerate(board):
         for x, tile in enumerate(row):
-            if slidingpuzzle.EMPTY_TILE == tile:
+            if BLANK_TILE == tile:
                 continue
-            orig_y = (tile - 1) // w
-            orig_x = (tile - 1) % w
-            dist += abs(y - orig_y) + abs(x - orig_x)
+            goal_y, goal_x = get_goal_yx(h, w, tile)
+            dist += abs(y - goal_y) + abs(x - goal_x)
     return dist
 
 
@@ -178,3 +198,36 @@ def random_distance(board: Board) -> int:
         A random distance that is consistent for a given board state
     """
     return hash(freeze_board(board))
+
+
+def relaxed_adjacency_distance(board: Board) -> int:
+    r"""
+    Repeatedly swap misplaced tiles into their goal positions, ignoring legal board
+    distance and rules. We repeat this process until the board is solved. This
+    heuristic is a slight improvement to Hamming distance.
+
+    Args:
+        board: The board
+
+    Returns:
+        The estimated distance to goal.
+    """
+    h, w = board.shape
+    board = np.copy(board)
+    dist = 0
+
+    def swap_first_misplaced(blank_yx: tuple[int, int]) -> None:
+        for y in range(h):
+            for x in range(w):
+                if get_goal_yx(h, w, board[y][x]) != (y, x):
+                    return swap_tiles(board, (y, x), blank_yx)
+
+    while not is_solved(board):
+        blank_yx = find_blank(board)
+        if blank_yx == get_goal_yx(h, w, BLANK_TILE):
+            swap_first_misplaced(blank_yx)
+        else:
+            swap_tiles(board, get_goal_tile(h, w, blank_yx), blank_yx)
+        dist += 1
+
+    return dist
